//---------------------------------------------------------------------------
//	Conditional Filter for VirtualDub - implements simple branching
//	Copyright (C) 2001-2003 Dmitri Schamschurko
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//      The author can be contacted at:
//      Dmitri Schamschurko
//      dc1@tcs.inf.tu-dresden.de
//
//      Version 1.0     Oktober 2001
//      Version 2.0     March 2003
//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "f_ConditionLogic.h"

//---------------------------------------------------------------------------

bool UseLuma709 = true;
bool UseSaturationHLS = true;
long CExecutionContext::lStamp = -1;

class CLookupTable {
  public:
    long Red[256];
    long Green[256];
    long Blue[256];

  protected:
    __fastcall CLookupTable(double r, double g)
    {
      for (int i = 0; i < 255; ++i)
      {
        Red[i]   = r * i;
        Green[i] = g * i;
        Blue[i]  = i - Red[i] - Green[i];
      }
    }
};

class CLuma601: public CLookupTable {
  public:
//    Y = 0.298912 R + 0.586611 G + 0.114478 B
    __fastcall CLuma601() : CLookupTable(0.298912, 0.586611) {}
} Luma601Tab;
class CLuma709: public CLookupTable {
  public:
//    Y = 0.212671 R + 0.715160 G + 0.072169 B
    __fastcall CLuma709() : CLookupTable(0.212671, 0.715160) {}
} Luma709Tab;
class CGray: public CLookupTable {
  public:
//    Y = 0.222015 R + 0.706655 G + 0.071330 B
    __fastcall CGray() : CLookupTable(0.222015, 0.706655) {}
} GrayTab;

__fastcall CMinMaxAvg::CMinMaxAvg()
{ lAvgCnt = 0; }

__fastcall CExecutionContext::CExecutionContext(const FilterActivation *vfa, long vFlags)
{
  CurrentStamp = lStamp++;
  Flags = vFlags;
  fa = vfa;
}

__fastcall CExecutionContext::~CExecutionContext()
{
}

void __fastcall CExecutionContext::ForceNeedAnything()
{
  if (lStamp != CurrentStamp)
    Flags = -1;
}

#ifndef max
#define max(a,b) (((a) > (b)) ? (a) : (b))
#endif
#ifndef min
#define min(a,b) (((a) < (b)) ? (a) : (b))
#endif

#define max3(a,b,c) max(a, max(b, c))
#define min3(a,b,c) min(a, min(b, c))
void __fastcall CExecutionContext::Compute()
{
  CurrentStamp = lStamp;

  PixDim w, h;
  Pixel32 *src;

  src = (Pixel32 *)fa->src.data;
  h = fa->src.h;

  bool NeedLuma601    = Flags & NEED_LUMA601;
  bool NeedLuma709    = Flags & NEED_LUMA709;
  bool NeedGray       = Flags & NEED_GRAY;
  bool NeedBrightness = Flags & NEED_SAT_BRIGHTNESS;
  bool NeedLightness  = Flags & NEED_SAT_LIGHTNESS;

  do {
    w = fa->src.w;

    do {
      Pixel32 v = *src;
      BYTE Red   = (v >> 16) & 0xff;
      BYTE Green = (v >> 8)  & 0xff;
      BYTE Blue  =  v        & 0xff;

      if (NeedLuma601) {
        long vLuma = Luma601Tab.Red[Red] + Luma601Tab.Green[Green] + Luma601Tab.Blue[Blue];
        Luma601.AddValue(vLuma);
      }
      if (NeedLuma709) {
        long vLuma = Luma709Tab.Red[Red] + Luma709Tab.Green[Green] + Luma709Tab.Blue[Blue];
        Luma709.AddValue(vLuma);
      }
      if (NeedGray) {
        long vLuma = GrayTab.Red[Red] + GrayTab.Green[Green] + GrayTab.Blue[Blue];
        Gray.AddValue(vLuma);
      }

      if (NeedBrightness || NeedLightness)
      {
        BYTE MaxNum = max3(Red, Green, Blue);
        BYTE MinNum = min3(Red, Green, Blue);
        long MaxPlusMin = MaxNum + MinNum;
        long MaxMinusMin = MaxNum - MinNum;

        if (NeedBrightness) {
          Brightness.AddValue(MaxNum);
          SaturationHSB.AddValue((MaxMinusMin << 8) / (MaxNum + 1)); // +1 is from me
        }

        if (NeedLightness) {
          Lightness.AddValue(MaxPlusMin); // div by 2!

          if (MaxPlusMin > 255)
            SaturationHLS.AddValue((MaxMinusMin << 8) / (MaxPlusMin + 1)); // +1 is from me
          else
            SaturationHLS.AddValue((MaxMinusMin << 8) / (255 + 255 - MaxPlusMin + 1)); // +1 is from me
        }
      }

      ++src;

    } while(--w);

    src = (Pixel32 *)((char *)src + fa->src.modulo);
  } while(--h);
}

long __fastcall CExecutionContext::GetGray(ExecuteModifier Mode)
{
  if (lStamp != CurrentStamp) {
    Flags |= NEED_GRAY; Compute();
  } else if (! (Flags & NEED_GRAY)) {
    PixDim w, h;
    Pixel32 *src;

    src = (Pixel32 *)fa->src.data;
    h = fa->src.h;

    Flags |= NEED_GRAY;

    do {
      w = fa->src.w;

      do {
        Pixel32 v = *src;
        BYTE Red   = (v >> 16) & 0xff;
        BYTE Green = (v >> 8)  & 0xff;
        BYTE Blue  =  v        & 0xff;

        long vLuma = GrayTab.Red[Red] + GrayTab.Green[Green] + GrayTab.Blue[Blue];
        Gray.AddValue(vLuma);

        ++src;

      } while(--w);

      src = (Pixel32 *)((char *)src + fa->src.modulo);
    } while(--h);
  }
  return Gray.Get(Mode);
}

long __fastcall CExecutionContext::GetLuma601(ExecuteModifier Mode)
{
  if (lStamp != CurrentStamp) {
    Flags |= NEED_LUMA601; Compute();
  } else if (! (Flags & NEED_LUMA601)) {
    PixDim w, h;
    Pixel32 *src;

    src = (Pixel32 *)fa->src.data;
    h = fa->src.h;

    Flags |= NEED_LUMA601;

    do {
      w = fa->src.w;

      do {
        Pixel32 v = *src;
        BYTE Red   = (v >> 16) & 0xff;
        BYTE Green = (v >> 8)  & 0xff;
        BYTE Blue  =  v        & 0xff;

        long vLuma = Luma601Tab.Red[Red] + Luma601Tab.Green[Green] + Luma601Tab.Blue[Blue];
        Luma601.AddValue(vLuma);

        ++src;

      } while(--w);

      src = (Pixel32 *)((char *)src + fa->src.modulo);
    } while(--h);
  }
  return Luma601.Get(Mode);
}

long __fastcall CExecutionContext::GetLuma709(ExecuteModifier Mode)
{
  if (lStamp != CurrentStamp) {
    Flags |= NEED_LUMA709; Compute();
  } else if (! (Flags & NEED_LUMA709)) {
    PixDim w, h;
    Pixel32 *src;

    src = (Pixel32 *)fa->src.data;
    h = fa->src.h;

    Flags |= NEED_LUMA709;

    do {
      w = fa->src.w;

      do {
        Pixel32 v = *src;
        BYTE Red   = (v >> 16) & 0xff;
        BYTE Green = (v >> 8)  & 0xff;
        BYTE Blue  =  v        & 0xff;

        long vLuma = Luma709Tab.Red[Red] + Luma709Tab.Green[Green] + Luma709Tab.Blue[Blue];
        Luma709.AddValue(vLuma);

        ++src;

      } while(--w);

      src = (Pixel32 *)((char *)src + fa->src.modulo);
    } while(--h);
  }
  return Luma709.Get(Mode);
}

long __fastcall CExecutionContext::GetBrightness(ExecuteModifier Mode)
{
  if (lStamp != CurrentStamp) {
    Flags |= NEED_SAT_BRIGHTNESS; Compute();
  } else if (! (Flags & NEED_SAT_BRIGHTNESS)) {
    PixDim w, h;
    Pixel32 *src;

    src = (Pixel32 *)fa->src.data;
    h = fa->src.h;

    Flags |= NEED_SAT_BRIGHTNESS;

    do {
      w = fa->src.w;

      do {
        Pixel32 v = *src;
        BYTE Red   = (v >> 16) & 0xff;
        BYTE Green = (v >> 8)  & 0xff;
        BYTE Blue  =  v        & 0xff;

        BYTE MaxNum = max3(Red, Green, Blue);
        BYTE MinNum = min3(Red, Green, Blue);
        long MaxMinusMin = MaxNum - MinNum;

        Brightness.AddValue(MaxNum);
        SaturationHSB.AddValue((MaxMinusMin << 8) / (MaxNum + 1)); // +1 is from me

        ++src;

      } while(--w);

      src = (Pixel32 *)((char *)src + fa->src.modulo);
    } while(--h);
  }
  return Brightness.Get(Mode);
}

long __fastcall CExecutionContext::GetLightness(ExecuteModifier Mode)
{
  if (lStamp != CurrentStamp) {
    Flags |= NEED_SAT_LIGHTNESS; Compute();
  } else if (! (Flags & NEED_SAT_LIGHTNESS)) {
    PixDim w, h;
    Pixel32 *src;

    src = (Pixel32 *)fa->src.data;
    h = fa->src.h;

    Flags |= NEED_SAT_LIGHTNESS;

    do {
      w = fa->src.w;

      do {
        Pixel32 v = *src;
        BYTE Red   = (v >> 16) & 0xff;
        BYTE Green = (v >> 8)  & 0xff;
        BYTE Blue  =  v        & 0xff;

        BYTE MaxNum = max3(Red, Green, Blue);
        BYTE MinNum = min3(Red, Green, Blue);
        long MaxPlusMin = MaxNum + MinNum;
        long MaxMinusMin = MaxNum - MinNum;

        Lightness.AddValue(MaxPlusMin); // div by 2!

        if (MaxPlusMin > 255)
          SaturationHLS.AddValue((MaxMinusMin << 8) / (MaxPlusMin + 1)); // +1 is from me
        else
          SaturationHLS.AddValue((MaxMinusMin << 8) / (255 + 255 - MaxPlusMin + 1)); // +1 is from me

        ++src;

      } while(--w);

      src = (Pixel32 *)((char *)src + fa->src.modulo);
    } while(--h);
  }
  return Lightness.Get(Mode);
}

long __fastcall CExecutionContext::GetSaturationHLS(ExecuteModifier Mode)
{
  if (lStamp != CurrentStamp) {
    Flags |= NEED_SAT_LIGHTNESS; Compute();
  } else if (! (Flags & NEED_SAT_LIGHTNESS)) {
    PixDim w, h;
    Pixel32 *src;

    src = (Pixel32 *)fa->src.data;
    h = fa->src.h;

    Flags |= NEED_SAT_LIGHTNESS;

    do {
      w = fa->src.w;

      do {
        Pixel32 v = *src;
        BYTE Red   = (v >> 16) & 0xff;
        BYTE Green = (v >> 8)  & 0xff;
        BYTE Blue  =  v        & 0xff;

        BYTE MaxNum = max3(Red, Green, Blue);
        BYTE MinNum = min3(Red, Green, Blue);
        long MaxPlusMin = MaxNum + MinNum;
        long MaxMinusMin = MaxNum - MinNum;

        Lightness.AddValue(MaxPlusMin); // div by 2!

        if (MaxPlusMin > 255)
          SaturationHLS.AddValue((MaxMinusMin << 8) / (MaxPlusMin + 1)); // +1 is from me
        else
          SaturationHLS.AddValue((MaxMinusMin << 8) / (255 + 255 - MaxPlusMin + 1)); // +1 is from me

        ++src;

      } while(--w);

      src = (Pixel32 *)((char *)src + fa->src.modulo);
    } while(--h);
  }
  return SaturationHLS.Get(Mode);
}

long __fastcall CExecutionContext::GetSaturationHSB(ExecuteModifier Mode)
{
  if (lStamp != CurrentStamp) {
    Flags |= NEED_SAT_BRIGHTNESS; Compute();
  } else if (! (Flags & NEED_SAT_BRIGHTNESS)) {
    PixDim w, h;
    Pixel32 *src;

    src = (Pixel32 *)fa->src.data;
    h = fa->src.h;

    Flags |= NEED_SAT_BRIGHTNESS;

    do {
      w = fa->src.w;

      do {
        Pixel32 v = *src;
        BYTE Red   = (v >> 16) & 0xff;
        BYTE Green = (v >> 8)  & 0xff;
        BYTE Blue  =  v        & 0xff;

        BYTE MaxNum = max3(Red, Green, Blue);
        BYTE MinNum = min3(Red, Green, Blue);
        long MaxMinusMin = MaxNum - MinNum;

        Brightness.AddValue(MaxNum);
        SaturationHSB.AddValue((MaxMinusMin << 8) / (MaxNum + 1)); // +1 is from me

        ++src;

      } while(--w);

      src = (Pixel32 *)((char *)src + fa->src.modulo);
    } while(--h);
  }
  return SaturationHSB.Get(Mode);
}

long __fastcall CExecutionContext::GetExternal(ExternalFunction fun, ExecuteModifier Mode)
{
  return fun(&fa->src, Mode, lStamp);
}

long __fastcall CExecutionContext::GetFrame()
{
  return fa->pfsi->lCurrentFrame;
}

long __fastcall CExecutionContext::GetStamp()
{
  return fa->pfsi->lCurrentSourceFrame;
}

#pragma package(smart_init)
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// numeric values
//
__fastcall NP::NP(CNumeric* Numeric)
{
  pNumeric = Numeric;
}
__fastcall NP::~NP()
{
  if (pNumeric)
    delete pNumeric;
}
long  __fastcall NP::GetValue(CExecutionContext& Context) const
{
  return pNumeric->GetValue(Context);
}
long __fastcall NP::GetNeedFlags() const
{
  return pNumeric->GetNeedFlags();
}
//---------------------------------------------------------------------------
__fastcall NRandom::NRandom(CNumeric* Numeric)
{
  pNumeric = Numeric;
}
__fastcall NRandom::~NRandom()
{
  if (pNumeric)
    delete pNumeric;
}
long  __fastcall NRandom::GetValue(CExecutionContext& Context) const
{
  long v = pNumeric->GetValue(Context);
  return random(v);
}
long __fastcall NRandom::GetNeedFlags() const
{
  return pNumeric->GetNeedFlags();
}
//---------------------------------------------------------------------------
__fastcall NTwoArgs::NTwoArgs(CNumeric* Numeric1, CNumeric* Numeric2)
{
  pNumeric1 = Numeric1;
  pNumeric2 = Numeric2;
}
__fastcall NTwoArgs::~NTwoArgs()
{
  if (pNumeric1)
    delete pNumeric1;
  if (pNumeric2)
    delete pNumeric2;
}
long __fastcall NTwoArgs::GetNeedFlags() const
{
  return (pNumeric1?pNumeric1->GetNeedFlags():0) | (pNumeric2?pNumeric2->GetNeedFlags():0);
}
long __fastcall NPlus::GetValue(CExecutionContext& Context) const
{
  return pNumeric1->GetValue(Context) + pNumeric2->GetValue(Context);
}
long __fastcall NMinus::GetValue(CExecutionContext& Context) const
{
  if (pNumeric1)
    return pNumeric1->GetValue(Context) - pNumeric2->GetValue(Context);
  return  - pNumeric2->GetValue(Context);
}
long __fastcall NMinus::GetNeedFlags() const
{
  return (pNumeric1 ? pNumeric1->GetNeedFlags() : 0) | pNumeric2->GetNeedFlags();
}
long __fastcall NTimes::GetValue(CExecutionContext& Context) const
{
  return pNumeric1->GetValue(Context) * pNumeric2->GetValue(Context);
}
long __fastcall NDiv::GetValue(CExecutionContext& Context) const
{
  return pNumeric1->GetValue(Context) / pNumeric2->GetValue(Context);
}
long __fastcall NDivRest::GetValue(CExecutionContext& Context) const
{
  return pNumeric1->GetValue(Context) % pNumeric2->GetValue(Context);
}
//---------------------------------------------------------------------------
__fastcall NNumber::NNumber(long Number)
{
  lNumber = Number;
}
long __fastcall NNumber::GetValue(CExecutionContext& Context) const
{
  return lNumber;
}
//---------------------------------------------------------------------------
long __fastcall NFrame::GetValue(CExecutionContext& Context) const
{
  return Context.GetFrame();
}
//---------------------------------------------------------------------------
long __fastcall NStamp::GetValue(CExecutionContext& Context) const
{
  return Context.GetStamp();
}
//---------------------------------------------------------------------------
__fastcall NFunction::NFunction(ExecuteModifier Modifier)
{
  pCallback = 0;
  pName     = 0;
  eModifier = Modifier;
}
__fastcall NFunction::NFunction(const char* Name, ExternalFunction Callback, ExecuteModifier Modifier)
{
  pCallback = Callback;
  pName     = Name;
  eModifier = Modifier;
}
long  __fastcall NFunction::GetValue(CExecutionContext& Context) const
{
  return Context.GetExternal(pCallback, eModifier);
}
const char* __fastcall NFunction::GetFunctionName() const
{
  return pName;
}
//---------------------------------------------------------------------------
__fastcall NGray::NGray(ExecuteModifier Modifier)
  : NFunction(Modifier) {}
long __fastcall NGray::GetValue(CExecutionContext& Context) const
{
  return Context.GetGray(eModifier);
}
long __fastcall NGray::GetNeedFlags() const
{
  return NEED_GRAY;
}
//---------------------------------------------------------------------------
__fastcall NLuma::NLuma(ExecuteModifier Modifier)
  : NFunction(Modifier) {}
long __fastcall NLuma::GetValue(CExecutionContext& Context) const
{
  return UseLuma709 ? Context.GetLuma709(eModifier) : Context.GetLuma601(eModifier);
}
long __fastcall NLuma::GetNeedFlags() const
{
  return UseLuma709 ? NEED_LUMA709 : NEED_LUMA601;
}
//---------------------------------------------------------------------------
__fastcall NLuma601::NLuma601(ExecuteModifier Modifier)
  : NFunction(Modifier) {}
long __fastcall NLuma601::GetValue(CExecutionContext& Context) const
{
  return Context.GetLuma601(eModifier);
}
long __fastcall NLuma601::GetNeedFlags() const
{
  return NEED_LUMA601;
}
//---------------------------------------------------------------------------
__fastcall NLuma709::NLuma709(ExecuteModifier Modifier)
  : NFunction(Modifier) {}
long __fastcall NLuma709::GetValue(CExecutionContext& Context) const
{
  return Context.GetLuma709(eModifier);
}
long __fastcall NLuma709::GetNeedFlags() const
{
  return NEED_LUMA709;
}
//---------------------------------------------------------------------------
__fastcall NLightness::NLightness(ExecuteModifier Modifier)
  : NFunction(Modifier) {}
long __fastcall NLightness::GetValue(CExecutionContext& Context) const
{
  return Context.GetLightness(eModifier);
}
long __fastcall NLightness::GetNeedFlags() const
{
  return NEED_SAT_LIGHTNESS;
}
//---------------------------------------------------------------------------
__fastcall NBrightness::NBrightness(ExecuteModifier Modifier)
  : NFunction(Modifier) {}
long __fastcall NBrightness::GetValue(CExecutionContext& Context) const
{
  return Context.GetBrightness(eModifier);
}
long __fastcall NBrightness::GetNeedFlags() const
{
  return NEED_SAT_BRIGHTNESS;
}
//---------------------------------------------------------------------------
__fastcall NSaturation::NSaturation(ExecuteModifier Modifier)
  : NFunction(Modifier) {}
long __fastcall NSaturation::GetValue(CExecutionContext& Context) const
{
  return UseSaturationHLS ? Context.GetSaturationHLS(eModifier) : Context.GetSaturationHSB(eModifier);
}
long __fastcall NSaturation::GetNeedFlags() const
{
  return UseSaturationHLS ? NEED_SAT_LIGHTNESS : NEED_SAT_BRIGHTNESS;
}
//---------------------------------------------------------------------------
__fastcall NSaturationHSB::NSaturationHSB(ExecuteModifier Modifier)
  : NFunction(Modifier) {}
long __fastcall NSaturationHSB::GetValue(CExecutionContext& Context) const
{
  return Context.GetSaturationHSB(eModifier);
}
long __fastcall NSaturationHSB::GetNeedFlags() const
{
  return NEED_SAT_BRIGHTNESS;
}
//---------------------------------------------------------------------------
__fastcall NSaturationHLS::NSaturationHLS(ExecuteModifier Modifier)
  : NFunction(Modifier) {}
long __fastcall NSaturationHLS::GetValue(CExecutionContext& Context) const
{
  return Context.GetSaturationHLS(eModifier);
}
long __fastcall NSaturationHLS::GetNeedFlags() const
{
  return NEED_SAT_LIGHTNESS;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// logical formulas
//
//---------------------------------------------------------------------------
__fastcall FOneArg::FOneArg(CFormula* Formula)
{
  pFormula = Formula;
}
__fastcall FOneArg::~FOneArg()
{
  if (pFormula)
    delete pFormula;
}
long __fastcall FOneArg::GetNeedFlags() const
{
  return pFormula->GetNeedFlags();
}
//---------------------------------------------------------------------------
__fastcall FNTwoArgs::FNTwoArgs(CNumeric* Numeric1, CNumeric* Numeric2)
{
  pNumeric1 = Numeric1;
  pNumeric2 = Numeric2;
}
__fastcall FNTwoArgs::~FNTwoArgs()
{
  if (pNumeric1)
    delete pNumeric1;
  if (pNumeric2)
    delete pNumeric2;
}
long __fastcall FNTwoArgs::GetNeedFlags() const
{
  return (pNumeric1 ? pNumeric1->GetNeedFlags() : 0) | (pNumeric2 ? pNumeric2->GetNeedFlags() : 0);
}
//---------------------------------------------------------------------------
__fastcall FFTwoArgs::FFTwoArgs(CFormula* Formula1, CFormula* Formula2)
{
  pFormula1 = Formula1;
  pFormula2 = Formula2;
}
__fastcall FFTwoArgs::~FFTwoArgs()
{
  if (pFormula1)
    delete pFormula1;
  if (pFormula2)
    delete pFormula2;
}
long __fastcall FFTwoArgs::GetNeedFlags() const
{
  return pFormula1->GetNeedFlags() | pFormula2->GetNeedFlags();
}
//---------------------------------------------------------------------------
bool __fastcall FTrue::Evaluate(CExecutionContext& Context) const
{
  return true;
}
//---------------------------------------------------------------------------
bool __fastcall FFalse::Evaluate(CExecutionContext& Context) const
{
  return false;
}
//---------------------------------------------------------------------------
bool __fastcall FP::Evaluate(CExecutionContext& Context) const
{
  return pFormula->Evaluate(Context);
}
//---------------------------------------------------------------------------
bool __fastcall FNot::Evaluate(CExecutionContext& Context) const
{
  return ! pFormula->Evaluate(Context);
}
//---------------------------------------------------------------------------
bool __fastcall FOr::Evaluate(CExecutionContext& Context) const
{
  return (pFormula1->Evaluate(Context) || pFormula2->Evaluate(Context));
}
long __fastcall FOr::GetNeedFlags() const
{
  return pFormula1->GetNeedFlags(); // | pFormula2->GetNeedFlags();
}
//---------------------------------------------------------------------------
bool __fastcall FAnd::Evaluate(CExecutionContext& Context) const
{
  return (pFormula1->Evaluate(Context) && pFormula2->Evaluate(Context));
}
//---------------------------------------------------------------------------
bool __fastcall FFEq::Evaluate(CExecutionContext& Context) const
{
  return (pFormula1->Evaluate(Context) == pFormula2->Evaluate(Context));
}
//---------------------------------------------------------------------------
bool __fastcall FFNotEq::Evaluate(CExecutionContext& Context) const
{
  return (pFormula1->Evaluate(Context) != pFormula2->Evaluate(Context));
}
//---------------------------------------------------------------------------
bool __fastcall FNEq::Evaluate(CExecutionContext& Context) const
{
  return (pNumeric1->GetValue(Context) == pNumeric2->GetValue(Context));
}
//---------------------------------------------------------------------------
bool __fastcall FNNotEq::Evaluate(CExecutionContext& Context) const
{
  return (pNumeric1->GetValue(Context) != pNumeric2->GetValue(Context));
}
//---------------------------------------------------------------------------
bool __fastcall FNLess::Evaluate(CExecutionContext& Context) const
{
  return (pNumeric1->GetValue(Context) < pNumeric2->GetValue(Context));
}
//---------------------------------------------------------------------------
bool __fastcall FNGreater::Evaluate(CExecutionContext& Context) const
{
  return (pNumeric1->GetValue(Context) > pNumeric2->GetValue(Context));
}
//---------------------------------------------------------------------------
bool __fastcall FNLessEq::Evaluate(CExecutionContext& Context) const
{
  return (pNumeric1->GetValue(Context) <= pNumeric2->GetValue(Context));
}
//---------------------------------------------------------------------------
bool __fastcall FNGreaterEq::Evaluate(CExecutionContext& Context) const
{
  return (pNumeric1->GetValue(Context) >= pNumeric2->GetValue(Context));
}
//---------------------------------------------------------------------------
__fastcall FFrame::FFrame(CNumeric* Eq)
  : FNTwoArgs(0, Eq)
{
  PlainNumbers = false;
}
__fastcall FFrame::FFrame(CNumeric* Up, CNumeric* To)
  : FNTwoArgs(Up, To)
{
  PlainNumbers = false;
}
__fastcall FFrame::FFrame(long Eq)
  : FNTwoArgs(0, 0)
{
  PlainNumbers = true;
  lUp = lTo = Eq;
}
__fastcall FFrame::FFrame(long Up, long To)
  : FNTwoArgs(0, 0)
{
  PlainNumbers = true;
  lUp = Up;
  lTo = To;
}
bool __fastcall FFrame::Evaluate(CExecutionContext& Context) const
{
  NFrame Val;
  long lVal = Val.NFrame::GetValue(Context);
  if (PlainNumbers)
    return (lUp <= lVal) && (lVal <= lTo);
  if (pNumeric1)
    return (pNumeric1->GetValue(Context) <= lVal) && (lVal <= pNumeric2->GetValue(Context));
  else
    return pNumeric2->GetValue(Context) == lVal;
}
//---------------------------------------------------------------------------
__fastcall FStamp::FStamp(CNumeric* Eq)
  : FNTwoArgs(0, Eq)
{
  PlainNumbers = false;
}
__fastcall FStamp::FStamp(CNumeric* Up, CNumeric* To)
  : FNTwoArgs(Up, To)
{
  PlainNumbers = false;
}
__fastcall FStamp::FStamp(long Eq)
  : FNTwoArgs(0, 0)
{
  PlainNumbers = true;
  lUp = lTo = Eq;
}
__fastcall FStamp::FStamp(long Up, long To)
  : FNTwoArgs(0, 0)
{
  PlainNumbers = true;
  lUp = Up;
  lTo = To;
}
bool __fastcall FStamp::Evaluate(CExecutionContext& Context) const
{
  NStamp Val;
  long lVal = Val.NStamp::GetValue(Context);
  if (PlainNumbers)
    return (lUp <= lVal) && (lVal <= lTo);
  if (pNumeric1)
    return (pNumeric1->GetValue(Context) <= lVal) && (lVal <= pNumeric2->GetValue(Context));
  else
    return pNumeric2->GetValue(Context) == lVal;
}
//---------------------------------------------------------------------------
#ifdef _DEBUG
CMyStrGuard __fastcall NP::TextOut()
{
  AnsiString t;

  t = AnsiString("( ") + pNumeric->TextOut().c_str() + " )";
  return CMyStrGuard(t.c_str());
}
CMyStrGuard __fastcall NRandom::TextOut()
{
  AnsiString t;

  t = AnsiString("rand( ") + pNumeric->TextOut().c_str() + " )";
  return CMyStrGuard(t.c_str());
}
CMyStrGuard __fastcall NTwoArgs::TextOut()
{
  AnsiString t;

  if (pNumeric1)
    t = AnsiString(pNumeric1->TextOut().c_str()) + " ";
  else
    t = "";

  t = t + GetOperator() + " " + pNumeric2->TextOut().c_str();
  return CMyStrGuard(t.c_str());
}
CMyStrGuard __fastcall NNumber::TextOut()
{
  AnsiString t;

  t = AnsiString(lNumber);
  return CMyStrGuard(t.c_str());
}
CMyStrGuard __fastcall NFunction::TextOut()
{
  AnsiString t;

  switch (eModifier) {
    case eExAvg:
      t = "avg("; break;
    case eExMax:
      t = "avg("; break;
    case eExMin:
      t = "avg("; break;
    default:
      t = "???(";
  }
  t += AnsiString(GetFunctionName()) + ")";
  return CMyStrGuard(t.c_str());
}
CMyStrGuard __fastcall FP::TextOut()
{
  AnsiString t;

  t = AnsiString("( ") + pFormula->TextOut().c_str() + " )";
  return CMyStrGuard(t.c_str());
}
CMyStrGuard __fastcall FNot::TextOut()
{
  AnsiString t;

  t = AnsiString("not ") + pFormula->TextOut().c_str();
  return CMyStrGuard(t.c_str());
}
CMyStrGuard __fastcall FNTwoArgs::TextOut()
{
  AnsiString t;

  if (pNumeric1)
    t = AnsiString(pNumeric1->TextOut().c_str()) + " ";
  else
    t = "";

  t = t + GetOperator() + " " + pNumeric2->TextOut().c_str();
  return CMyStrGuard(t.c_str());
}
CMyStrGuard __fastcall FFTwoArgs::TextOut()
{
  AnsiString t;

  t = AnsiString(pFormula1->TextOut().c_str()) + " " + GetOperator() + " " + pFormula2->TextOut().c_str();
  return CMyStrGuard(t.c_str());
}
CMyStrGuard __fastcall FFrame::TextOut()
{
  AnsiString t;

  AnsiString n1, n2;
  if (PlainNumbers) {

    if (lUp == lTo) {
      n1 = "";
      n2 = AnsiString(lTo);
    } else {
      n1 = lUp;
      n2 = AnsiString("..") + AnsiString(lTo);
    }

  } else {

    if (pNumeric1) {
      n1 = pNumeric1->TextOut().c_str();
      n2 = AnsiString("..") + pNumeric2->TextOut().c_str();
    } else {
      n1 = "";
      n2 = pNumeric2->TextOut().c_str();
    }

  }

  t = AnsiString("frame(") + n1 + n2 + ")";
  return CMyStrGuard(t.c_str());
}
CMyStrGuard __fastcall FStamp::TextOut()
{
  AnsiString t;

  AnsiString n1, n2;
  if (PlainNumbers) {

    if (lUp == lTo) {
      n1 = "";
      n2 = AnsiString(lTo);
    } else {
      n1 = lUp;
      n2 = AnsiString("..") + AnsiString(lTo);
    }

  } else {

    if (pNumeric1) {
      n1 = pNumeric1->TextOut().c_str();
      n2 = AnsiString("..") + pNumeric2->TextOut().c_str();
    } else {
      n1 = "";
      n2 = pNumeric2->TextOut().c_str();
    }

  }

  t = AnsiString("stamp(") + n1 + n2 + ")";
  return CMyStrGuard(t.c_str());
}
#endif

