//---------------------------------------------------------------------------
//	Conditional Filter for VirtualDub - implements simple branching
//	Copyright (C) 2001-2003 Dmitri Schamschurko
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//      The author can be contacted at:
//      Dmitri Schamschurko
//      dc1@tcs.inf.tu-dresden.de
//
//      Version 1.0     Oktober 2001
//      Version 2.0     March 2003
//---------------------------------------------------------------------------

#ifndef f_ConditionLogicH
#define f_ConditionLogicH

#include <condfun.h>
#include <filter.h>
//---------------------------------------------------------------------------
// Basic classes
//
enum ExecuteModifier {
  eExAvg = ExAvg,
  eExMin = ExMin,
  eExMax = ExMax
};

class CMinMaxAvg {
  private:
    __int64 lAvgSum;
    long    lAvgCnt;
    long    lMin;
    long    lMax;

  private:
    bool    NewValue;

  public:
    __fastcall CMinMaxAvg();

    long inline Get(ExecuteModifier Modifier) const
    {
      if (Modifier == eExMin) return lMin;
      if (Modifier == eExMax) return lMax;
      return lAvgSum / lAvgCnt;
    }
    void inline AddValue(long Value)
    {
      if (lAvgCnt++) {
        lAvgSum += Value;
        if (lMax < Value)
          lMax = Value;
        else if (lMin > Value)
          lMin = Value;
      } else {
        lAvgSum =
        lMin = lMax = Value;
      }
    }
};

class CExecutionContext {
  private:
    static long lStamp;
    long CurrentStamp;
    long Flags;
    const FilterActivation* fa;

  private:
    CMinMaxAvg Luma601;
    CMinMaxAvg Luma709;
    CMinMaxAvg Gray;
    CMinMaxAvg Brightness;
    CMinMaxAvg Lightness;
    CMinMaxAvg SaturationHSB;
    CMinMaxAvg SaturationHLS;

    void __fastcall Compute();

  public:
    __fastcall CExecutionContext(const FilterActivation *vfa, long vFlags = -1);
    __fastcall ~CExecutionContext();
    void __fastcall ForceNeedAnything();

    long __fastcall GetGray(ExecuteModifier Mode);
    long __fastcall GetLuma601(ExecuteModifier Mode);
    long __fastcall GetLuma709(ExecuteModifier Mode);
    long __fastcall GetBrightness(ExecuteModifier Mode);
    long __fastcall GetLightness(ExecuteModifier Mode);
    long __fastcall GetSaturationHLS(ExecuteModifier Mode);
    long __fastcall GetSaturationHSB(ExecuteModifier Mode);
    long __fastcall GetExternal(ExternalFunction fun, ExecuteModifier Mode);
    long __fastcall GetFrame();
    long __fastcall GetStamp();
};

const int NEED_LUMA601        = 0x0001;
const int NEED_LUMA709        = 0x0002;
const int NEED_GRAY           = 0x0004;
const int NEED_SAT_LIGHTNESS  = 0x0008;
const int NEED_SAT_BRIGHTNESS = 0x0010;

#ifdef _DEBUG
class CMyStrGuard {
  protected:
    const char* pStr;

  public:
    CMyStrGuard(const char* aStr)
    {
      char* p = new char[strlen(aStr) + 1];
      strcpy(p, aStr);
      pStr = p;
    }
    CMyStrGuard(const CMyStrGuard& Other)
    {
      char* p = new char[strlen(Other.pStr) + 1];
      strcpy(p, Other.pStr);
      pStr = p;
    }
    ~CMyStrGuard()
    {
      delete [] pStr;
    }
    CMyStrGuard& operator= (const CMyStrGuard& Other)
    {
      if (this != &Other) {
        char* p = new char[strlen(Other.pStr) + 1];
        strcpy(p, Other.pStr);

        delete [] pStr;
        pStr = p;
      }
      return *this;
    }
    const char* c_str() { return pStr; }
};
#endif

// This class is used only to make the chain which can handle
// parse errors. In that case we cannot call the standard destructor!
//__declspec()
class CChainElement {
  private:
    CChainElement(const CChainElement&);
    CChainElement& operator=(const CChainElement&);

  public:
    CChainElement* Next;

  public:
    __fastcall CChainElement()
    { Next = 0; }
    virtual __fastcall ~CChainElement()
    { }
    virtual void __fastcall Clear(bool ClearLocal = false)
    { if (Next) Next->Clear(ClearLocal); delete this; }
#ifdef _DEBUG
    virtual CMyStrGuard __fastcall TextOut() = 0;
#endif
};

class CNumeric: public CChainElement {
  public:
    virtual long __fastcall GetNeedFlags() const { return 0; }
    virtual long __fastcall GetValue(CExecutionContext& Context) const = 0;
};

class CFormula: public CChainElement {
  public:
    virtual long __fastcall GetNeedFlags() const { return 0; }
    virtual bool __fastcall Evaluate(CExecutionContext& Context) const = 0;
};

//---------------------------------------------------------------------------
// numeric values
//
class NP: public CNumeric {
  private:
    CNumeric* pNumeric;

  public:
    __fastcall NP(CNumeric* Numeric);
    virtual __fastcall ~NP();
    virtual long  __fastcall GetValue(CExecutionContext& Context) const;
    virtual long __fastcall GetNeedFlags() const;
    virtual void __fastcall Clear(bool ClearLocal = false)
    { if (Next)       Next->Clear(ClearLocal);
      if (ClearLocal) pNumeric = 0;
      delete this; }
#ifdef _DEBUG
    virtual CMyStrGuard __fastcall TextOut();
#endif
};
class NRandom: public CNumeric {
  private:
    CNumeric* pNumeric;

  public:
    __fastcall NRandom(CNumeric* Numeric);
    virtual __fastcall ~NRandom();
    virtual long  __fastcall GetValue(CExecutionContext& Context) const;
    virtual long __fastcall GetNeedFlags() const;
    virtual void __fastcall Clear(bool ClearLocal = false)
    { if (Next)       Next->Clear(ClearLocal);
      if (ClearLocal) pNumeric = 0;
      delete this; }
#ifdef _DEBUG
    virtual CMyStrGuard __fastcall TextOut();
#endif
};
class NTwoArgs: public CNumeric {
  protected:
    CNumeric* pNumeric1;
    CNumeric* pNumeric2;

  public:
    __fastcall NTwoArgs(CNumeric* Numeric1, CNumeric* Numeric2);
    virtual __fastcall ~NTwoArgs();
    virtual long __fastcall GetNeedFlags() const;
    virtual void __fastcall Clear(bool ClearLocal = false)
    { if (Next)       Next->Clear(ClearLocal);
      if (ClearLocal) { pNumeric1 = 0; pNumeric2 = 0; }
      delete this; }
    virtual const char* __fastcall GetOperator() = 0;
#ifdef _DEBUG
    virtual CMyStrGuard __fastcall TextOut();
#endif
};
class NPlus: public NTwoArgs {
  public:
    __fastcall NPlus(CNumeric* n1, CNumeric* n2)
      : NTwoArgs(n1, n2) {}
    virtual long __fastcall GetValue(CExecutionContext& Context) const;
    virtual const char* __fastcall GetOperator() { return "+"; }
};
class NMinus: public NTwoArgs {
  public:
    __fastcall NMinus(CNumeric* n1, CNumeric* n2)
      : NTwoArgs(n1, n2) {}
    __fastcall NMinus(CNumeric* Numeric)
      : NTwoArgs(0, Numeric) {}
    virtual long __fastcall GetValue(CExecutionContext& Context) const;
    virtual long __fastcall GetNeedFlags() const;
    virtual const char* __fastcall GetOperator() { return "-"; }
};
class NTimes: public NTwoArgs {
  public:
    __fastcall NTimes(CNumeric* n1, CNumeric* n2)
      : NTwoArgs(n1, n2) {}
    virtual long __fastcall GetValue(CExecutionContext& Context) const;
    virtual const char* __fastcall GetOperator() { return "*"; }
};
class NDiv: public NTwoArgs {
  public:
    __fastcall NDiv(CNumeric* n1, CNumeric* n2)
      : NTwoArgs(n1, n2) {}
    virtual long __fastcall GetValue(CExecutionContext& Context) const;
    virtual const char* __fastcall GetOperator() { return "/"; }
};
class NDivRest: public NTwoArgs {
  public:
    __fastcall NDivRest(CNumeric* n1, CNumeric* n2)
      : NTwoArgs(n1, n2) {}
    virtual long __fastcall GetValue(CExecutionContext& Context) const;
    virtual const char* __fastcall GetOperator() { return "%"; }
};
class NNumber: public CNumeric {
  private:
    long lNumber;

  public:
    __fastcall NNumber(long Number);
    virtual long __fastcall GetValue(CExecutionContext& Context) const;
#ifdef _DEBUG
    virtual CMyStrGuard __fastcall TextOut();
#endif
};
class NFrame: public CNumeric {
  public:
    virtual long __fastcall GetValue(CExecutionContext& Context) const;
#ifdef _DEBUG
    virtual CMyStrGuard __fastcall TextOut() { return "frame"; }
#endif
};
class NStamp: public CNumeric {
  public:
    virtual long __fastcall GetValue(CExecutionContext& Context) const;
#ifdef _DEBUG
    virtual CMyStrGuard __fastcall TextOut() { return "stamp"; }
#endif
};
class NFunction: public CNumeric {
  private:
    ExternalFunction pCallback;
    const char*      pName;

  public:
    ExecuteModifier eModifier;

  protected:
    __fastcall NFunction(ExecuteModifier Modifier);

  public:
    __fastcall NFunction(const char* Name, ExternalFunction Callback, ExecuteModifier Modifier = eExAvg);
    virtual long  __fastcall GetValue(CExecutionContext& Context) const;
    virtual const char* __fastcall GetFunctionName() const;
#ifdef _DEBUG
    virtual CMyStrGuard __fastcall TextOut();
#endif
};
class NGray: public NFunction {
  public:
    __fastcall NGray(ExecuteModifier Modifier = eExAvg);
    virtual long  __fastcall GetValue(CExecutionContext& Context) const;
    virtual const char* __fastcall GetFunctionName() const { return "gray"; }
    virtual long __fastcall GetNeedFlags() const;
};
class NLuma: public NFunction {
  public:
    __fastcall NLuma(ExecuteModifier Modifier = eExAvg);
    virtual long  __fastcall GetValue(CExecutionContext& Context) const;
    virtual const char* __fastcall GetFunctionName() const { return "luma"; }
    virtual long __fastcall GetNeedFlags() const;
};
class NLuma601: public NFunction {
  public:
    __fastcall NLuma601(ExecuteModifier Modifier = eExAvg);
    virtual long  __fastcall GetValue(CExecutionContext& Context) const;
    virtual const char* __fastcall GetFunctionName() const { return "luma601"; }
    virtual long __fastcall GetNeedFlags() const;
};
class NLuma709: public NFunction {
  public:
    __fastcall NLuma709(ExecuteModifier Modifier = eExAvg);
    virtual long  __fastcall GetValue(CExecutionContext& Context) const;
    virtual const char* __fastcall GetFunctionName() const { return "luma709"; }
    virtual long __fastcall GetNeedFlags() const;
};
class NLightness: public NFunction {
  public:
    __fastcall NLightness(ExecuteModifier Modifier = eExAvg);
    virtual long  __fastcall GetValue(CExecutionContext& Context) const;
    virtual const char* __fastcall GetFunctionName() const { return "lightness"; }
    virtual long __fastcall GetNeedFlags() const;
};
class NBrightness: public NFunction {
  public:
    __fastcall NBrightness(ExecuteModifier Modifier = eExAvg);
    virtual long  __fastcall GetValue(CExecutionContext& Context) const;
    virtual const char* __fastcall GetFunctionName() const { return "brightness"; }
    virtual long __fastcall GetNeedFlags() const;
};
class NSaturation: public NFunction {
  public:
    __fastcall NSaturation(ExecuteModifier Modifier = eExAvg);
    virtual long  __fastcall GetValue(CExecutionContext& Context) const;
    virtual const char* __fastcall GetFunctionName() const { return "saturation"; }
    virtual long __fastcall GetNeedFlags() const;
};
class NSaturationHSB: public NFunction {
  public:
    __fastcall NSaturationHSB(ExecuteModifier Modifier = eExAvg);
    virtual long  __fastcall GetValue(CExecutionContext& Context) const;
    virtual const char* __fastcall GetFunctionName() const { return "saturationHSB"; }
    virtual long __fastcall GetNeedFlags() const;
};
class NSaturationHLS: public NFunction {
  public:
    __fastcall NSaturationHLS(ExecuteModifier Modifier = eExAvg);
    virtual long  __fastcall GetValue(CExecutionContext& Context) const;
    virtual const char* __fastcall GetFunctionName() const { return "saturationHLS"; }
    virtual long __fastcall GetNeedFlags() const;
};

//---------------------------------------------------------------------------
// logical formulas
//
class FOneArg: public CFormula {
  protected:
    CFormula* pFormula;

  public:
    __fastcall FOneArg(CFormula* Formula);
    virtual __fastcall ~FOneArg();
    virtual long __fastcall GetNeedFlags() const;
    virtual void __fastcall Clear(bool ClearLocal = false)
    { if (Next)       Next->Clear(ClearLocal);
      if (ClearLocal) pFormula = 0;
      delete this; }
};
class FNTwoArgs: public CFormula {
  protected:
    CNumeric* pNumeric1;
    CNumeric* pNumeric2;

  public:
    __fastcall FNTwoArgs(CNumeric* Numeric1, CNumeric* Numeric2);
    virtual __fastcall ~FNTwoArgs();
    virtual long __fastcall GetNeedFlags() const;
    virtual void __fastcall Clear(bool ClearLocal = false)
    { if (Next)       Next->Clear(ClearLocal);
      if (ClearLocal) { pNumeric1 = 0; pNumeric2 = 0; }
      delete this; }
    virtual const char* __fastcall GetOperator() = 0;
#ifdef _DEBUG
    virtual CMyStrGuard __fastcall TextOut();
#endif
};
class FFTwoArgs: public CFormula {
  protected:
    CFormula* pFormula1;
    CFormula* pFormula2;

  public:
    __fastcall FFTwoArgs(CFormula* Formula1, CFormula* Formula2);
    virtual __fastcall ~FFTwoArgs();
    virtual long __fastcall GetNeedFlags() const;
    virtual void __fastcall Clear(bool ClearLocal = false)
    { if (Next)       Next->Clear(ClearLocal);
      if (ClearLocal) { pFormula1 = 0; pFormula2 = 0; }
      delete this; }
    virtual const char* __fastcall GetOperator() = 0;
#ifdef _DEBUG
    virtual CMyStrGuard __fastcall TextOut();
#endif
};
class FTrue: public CFormula {
  public:
    virtual bool __fastcall Evaluate(CExecutionContext& Context) const;
#ifdef _DEBUG
    virtual CMyStrGuard __fastcall TextOut() { return "true"; }
#endif
};
class FFalse: public CFormula {
  public:
    virtual bool __fastcall Evaluate(CExecutionContext& Context) const;
#ifdef _DEBUG
    virtual CMyStrGuard __fastcall TextOut() { return "false"; }
#endif
};
class FP: public FOneArg {
  public:
    __fastcall FP(CFormula* f)
      : FOneArg(f) {}
    virtual bool __fastcall Evaluate(CExecutionContext& Context) const;
#ifdef _DEBUG
    virtual CMyStrGuard __fastcall TextOut();
#endif
};
class FNot: public FOneArg {
  public:
    __fastcall FNot(CFormula* f)
      : FOneArg(f) {}
    virtual bool __fastcall Evaluate(CExecutionContext& Context) const;
#ifdef _DEBUG
    virtual CMyStrGuard __fastcall TextOut();
#endif
};
class FOr: public FFTwoArgs {
  public:
    __fastcall FOr(CFormula* f1, CFormula* f2)
      : FFTwoArgs(f1, f2) {}
    virtual long __fastcall GetNeedFlags() const;
    virtual bool __fastcall Evaluate(CExecutionContext& Context) const;
    virtual const char* __fastcall GetOperator() { return "or"; }
};
class FAnd: public FFTwoArgs {
  public:
    __fastcall FAnd(CFormula* f1, CFormula* f2)
      : FFTwoArgs(f1, f2) {}
    virtual bool __fastcall Evaluate(CExecutionContext& Context) const;
    virtual const char* __fastcall GetOperator() { return "and"; }
};
class FFEq: public FFTwoArgs {
  public:
    __fastcall FFEq(CFormula* f1, CFormula* f2)
      : FFTwoArgs(f1, f2) {}
    virtual bool __fastcall Evaluate(CExecutionContext& Context) const;
    virtual const char* __fastcall GetOperator() { return "="; }
};
class FFNotEq: public FFTwoArgs {
  public:
    __fastcall FFNotEq(CFormula* f1, CFormula* f2)
      : FFTwoArgs(f1, f2) {}
    virtual bool __fastcall Evaluate(CExecutionContext& Context) const;
    virtual const char* __fastcall GetOperator() { return "!="; }
};
class FNEq: public FNTwoArgs {
  public:
    __fastcall FNEq(CNumeric* n1, CNumeric* n2)
      : FNTwoArgs(n1, n2) {}
    virtual bool __fastcall Evaluate(CExecutionContext& Context) const;
    virtual const char* __fastcall GetOperator() { return "="; }
};
class FNNotEq: public FNTwoArgs {
  public:
    __fastcall FNNotEq(CNumeric* n1, CNumeric* n2)
      : FNTwoArgs(n1, n2) {}
    virtual bool __fastcall Evaluate(CExecutionContext& Context) const;
    virtual const char* __fastcall GetOperator() { return "!="; }
};
class FNLess: public FNTwoArgs {
  public:
    __fastcall FNLess(CNumeric* n1, CNumeric* n2)
      : FNTwoArgs(n1, n2) {}
    virtual bool __fastcall Evaluate(CExecutionContext& Context) const;
    virtual const char* __fastcall GetOperator() { return "<"; }
};
class FNGreater: public FNTwoArgs {
  public:
    __fastcall FNGreater(CNumeric* n1, CNumeric* n2)
      : FNTwoArgs(n1, n2) {}
    virtual bool __fastcall Evaluate(CExecutionContext& Context) const;
    virtual const char* __fastcall GetOperator() { return ">"; }
};
class FNLessEq: public FNTwoArgs {
  public:
    __fastcall FNLessEq(CNumeric* n1, CNumeric* n2)
      : FNTwoArgs(n1, n2) {}
    virtual bool __fastcall Evaluate(CExecutionContext& Context) const;
    virtual const char* __fastcall GetOperator() { return "<="; }
};
class FNGreaterEq: public FNTwoArgs {
  public:
    __fastcall FNGreaterEq(CNumeric* n1, CNumeric* n2)
      : FNTwoArgs(n1, n2) {}
    virtual bool __fastcall Evaluate(CExecutionContext& Context) const;
    virtual const char* __fastcall GetOperator() { return ">="; }
};
class FFrame: public FNTwoArgs {
  private:
    bool      PlainNumbers;
    long      lUp;
    long      lTo;

  public:
    __fastcall FFrame(CNumeric* Eq);
    __fastcall FFrame(CNumeric* Up, CNumeric* To);
    __fastcall FFrame(long Eq);
    __fastcall FFrame(long Up, long To);
    virtual bool __fastcall Evaluate(CExecutionContext& Context) const;
    virtual const char* __fastcall GetOperator() { return "..."; }
#ifdef _DEBUG
    virtual CMyStrGuard __fastcall TextOut();
#endif
};
class FStamp: public FNTwoArgs {
  private:
    bool      PlainNumbers;
    long      lUp;
    long      lTo;

  public:
    __fastcall FStamp(CNumeric* Eq);
    __fastcall FStamp(CNumeric* Up, CNumeric* To);
    __fastcall FStamp(long Eq);
    __fastcall FStamp(long Up, long To);
    virtual bool __fastcall Evaluate(CExecutionContext& Context) const;
    virtual const char* __fastcall GetOperator() { return "..."; }
#ifdef _DEBUG
    virtual CMyStrGuard __fastcall TextOut();
#endif
};
//---------------------------------------------------------------------------
#endif

