//---------------------------------------------------------------------------
//	Conditional Filter for VirtualDub - implements simple branching
//	Copyright (C) 2001-2003 Dmitri Schamschurko
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//      The author can be contacted at:
//      Dmitri Schamschurko
//      dc1@tcs.inf.tu-dresden.de
//
//      Version 1.0     Oktober 2001
//      Version 2.0     March 2003
//
//      The code in this file uses slightly modified code from the source
//      of VirtualDub v 1.5.1 (http://www.virtualdub.org/)
//      copyrighted by Avery Lee
//
//      Copyright note from that file:
//---------------------------------------------------------------------------
//	VirtualDub - Video processing and capture application
//	Copyright (C) 1998-2001 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "f_VD15xProxy.h"
#include "f_ConditionData.h"

extern const FilterFunctions* gmy_filterFuncs;
extern int                    gmy_VIRTUALDUB_FILTERDEF_VERSION;

#include "vd2/system/vdstring.h"
#include "vd2/system/filesys.h"
#include "vd2/system/error.h"
//---------------------------------------------------------------------------

#pragma package(smart_init)

namespace CondVD15ns {

// code by dima
// proxy replacement for ListNode2
template <class T>
class ListNode2 {
public:
	T *next, *prev;
};

///////////////////////////////////////////////////////////////////////////
//
//	FilterModuleInstance
//
///////////////////////////////////////////////////////////////////////////

class FilterModuleInstance : public ListNode2<FilterModuleInstance> {
public:
	FilterModuleInstance(const VDStringW& filename);
	~FilterModuleInstance();

	void AttachToModule();
	void ReleaseModule();

	const VDStringW& GetName() const { return mFilename; }
	const FilterModule& GetInfo() const { return mModuleInfo; }

	bool IsLoaded() const { return mModuleInfo.hInstModule != 0; }

protected:
	void Load();
	void Unload();

	VDStringW		mFilename;
	FilterModule	mModuleInfo;
	VDAtomicInt		mRefCount;
};

FilterModuleInstance::FilterModuleInstance(const VDStringW& filename)
	: mFilename(filename)
	, mRefCount(0)
{
	memset(&mModuleInfo, 0, sizeof mModuleInfo);
}

FilterModuleInstance::~FilterModuleInstance() {
	Unload();
}

void FilterModuleInstance::AttachToModule() {
	if (!mRefCount.postinc())
		try {
			Load();
		} catch(...) {
			--mRefCount;
			throw;
		}
}

void FilterModuleInstance::ReleaseModule() {
	VDASSERT(mRefCount > 0);
                                    
	if (mRefCount.dectestzero())
		Unload();
}

void FilterModuleInstance::Load() {
/*
  dout(__FUNC__ " 1");

  dout(__FUNC__ " 1 prev: %p", prev);
  dout(__FUNC__ " 1 next: %p", next);
  dout(__FUNC__ " 1 mi.prev: %p", mModuleInfo.prev);
  dout(__FUNC__ " 1 mi.next: %p", mModuleInfo.next);
  dout(__FUNC__ " 1 mi.inst: %p", mModuleInfo.hInstModule);
  dout(__FUNC__ " 1 mi.init: %p", mModuleInfo.initProc);
  dout(__FUNC__ " 1 ref_cnt: %d", (int)mRefCount);
  dout(__FUNC__ " 1 cond: %p", fd_conditional->module->hInstModule);
*/

	if (!mModuleInfo.hInstModule)
        {
		mModuleInfo.hInstModule = LoadLibrary(VDTextWToA(mFilename).c_str());

		const VDStringA name(VDTextWToA(VDFileSplitPathRight(mFilename)));

		if (!mModuleInfo.hInstModule)
			throw MyWin32Error("Cannot load filter module \"%s\": %%s", GetLastError());

		try {
			mModuleInfo.initProc   = (FilterModuleInitProc  )GetProcAddress(mModuleInfo.hInstModule, "VirtualdubFilterModuleInit2");
			mModuleInfo.deinitProc = (FilterModuleDeinitProc)GetProcAddress(mModuleInfo.hInstModule, "VirtualdubFilterModuleDeinit");

			if (!mModuleInfo.initProc) {
				void *fp = GetProcAddress(mModuleInfo.hInstModule, "VirtualdubFilterModuleInit");

				if (fp)
					throw MyError(
						"This filter was created for VirtualDub 1.1 or earlier, and is not compatible with version 1.2 or later. "
						"Please contact the author for an updated version.");

				fp = GetProcAddress(mModuleInfo.hInstModule, "NinaFilterModuleInit");

				if (fp)
					throw MyError("This filter will only work with VirtualDub 2.0 or above.");
			}

			if (!mModuleInfo.initProc || !mModuleInfo.deinitProc)
				throw MyError("Module \"%s\" does not contain VirtualDub filters.", name.c_str());

			int ver_hi = gmy_VIRTUALDUB_FILTERDEF_VERSION;
			int ver_lo = VIRTUALDUB_FILTERDEF_COMPATIBLE;

			if (mModuleInfo.initProc(&mModuleInfo, gmy_filterFuncs, ver_hi, ver_lo))
				throw MyError("Error initializing module \"%s\".",name.c_str());

			if (ver_hi < VIRTUALDUB_FILTERDEF_COMPATIBLE) {
				mModuleInfo.deinitProc(&mModuleInfo, gmy_filterFuncs);

				throw MyError(
					"This filter was created for an earlier, incompatible filter interface. As a result, it will not "
					"run correctly with this version of VirtualDub. Please contact the author for an updated version.");
			}

			if (ver_lo > VIRTUALDUB_FILTERDEF_VERSION) {
				mModuleInfo.deinitProc(&mModuleInfo, gmy_filterFuncs);

				throw MyError(
					"This filter uses too new of a filter interface!  You'll need to upgrade to a newer version of "
					"VirtualDub to use this filter."
					);
			}
		} catch(...) {
			FreeLibrary(mModuleInfo.hInstModule);
			mModuleInfo.hInstModule = NULL;
			throw;
		}
	}
}

void FilterModuleInstance::Unload() {
	if (mModuleInfo.hInstModule) {
		mModuleInfo.deinitProc(&mModuleInfo, gmy_filterFuncs);
		FreeLibrary(mModuleInfo.hInstModule);
		mModuleInfo.hInstModule = NULL;
	}
}

///////////////////////////////////////////////////////////////////////////
//
//	FilterDefinitionInstance
//
///////////////////////////////////////////////////////////////////////////

class FilterDefinitionInstance : public ListNode2<FilterDefinitionInstance> {
public:
	FilterDefinitionInstance(FilterModuleInstance *pfm);
	~FilterDefinitionInstance();

	void Assign(const FilterDefinition& def, int len);

	const FilterDefinition& Attach();
	void Detach();

	const FilterDefinition& GetDef() const { return mDef; }
	FilterModuleInstance *GetModule() const { return mpModule; }

	const VDStringA&	GetName() const { return mName; }
	const VDStringA&	GetAuthor() const { return mAuthor; }
	const VDStringA&	GetDescription() const { return mDescription; }

public:
	FilterModuleInstance *const mpModule;
	FilterDefinition mDef;
        // we use FilterDefinition from 1.4
        // in the newer version there are 4 new functions, which
        // are not yet included here
        void *string2Proc, *serializeProc, *deserializeProc, *copyProc;

	VDAtomicInt			mRefCount;
	VDStringA			mName;
	VDStringA			mAuthor;
	VDStringA			mDescription;
};

FilterDefinitionInstance::FilterDefinitionInstance(FilterModuleInstance *pfm)
	: mpModule(pfm)
	, mRefCount(0)
{
}

FilterDefinitionInstance::~FilterDefinitionInstance() {
	VDASSERT(mRefCount==0);
}

/*
void FilterDefinitionInstance::Assign(const FilterDefinition& def, int len) {
	memset(&mDef, 0, sizeof mDef);
	memcpy(&mDef, &def, std::min<size_t>(sizeof mDef, len));

	mName			= def.name;
	mAuthor			= def.maker ? def.maker : "(internal)";
	mDescription	= def.desc;

	mDef.module		= const_cast<FilterModule *>(&mpModule->GetInfo());
}
*/

const FilterDefinition& FilterDefinitionInstance::Attach() {
#ifdef _DEBUG
  dout("++++++++++++++++++");
  try {
    if (GetName().empty())
      dout("+ name is empty");
    else
      dout("+ name is %s", GetName().c_str());
  } __except(1) {
      dout("+ name EXCEPTION");
  }

  try {
    if (!mDef.name)
      dout("+ def name is empty");
    else
      dout("+ def name is %s", mDef.name);
  } __except(1) {
      dout("+ def name EXCEPTION");
  }

  try {
    if (mpModule) {
      dout("+ module is %ls", mpModule->GetName().c_str());
    } else {
      dout("+ module is empty");
    }
  } __except(1) {
      dout("+ def name EXCEPTION");
  }
  dout("-------------------");
#endif

	if (mpModule)
		mpModule->AttachToModule();

	++mRefCount;

dout("++ attached to   '%s'", mDef.name);

	return mDef;
}

void FilterDefinitionInstance::Detach() {
	VDASSERT(mRefCount.dec() >= 0);

dout("++ detached from '%s'", mDef.name);

	if (mpModule)
		mpModule->ReleaseModule();
}

};

CFilterDefinitionHelper::CFilterDefinitionHelper(FilterDefinition* fd)
{
  m_VD_15x_Mode = (!fd->next) && (!fd->prev);

#ifdef _DEBUG
  // test
  if (m_VD_15x_Mode) {
    dout("+++ vd 1.5.x mode");
  } else {
    dout("+++ vd 1.4.x mode");
  }

  FilterDefinition* fdtest;
  fdtest = GetPrev(fd);
  if (fdtest && (GetNext(fdtest) != fd)) {
    dout("+++ prev.next != id");
  }
  fdtest = GetNext(fd);
  if (fdtest && (GetPrev(fdtest) != fd)) {
    dout("+++ next.prev != id");
  }
#endif
}

FilterDefinition* CFilterDefinitionHelper::GetPrev(FilterDefinition* fd)
{
  if (!m_VD_15x_Mode) return fd->prev;

  CondVD15ns::FilterDefinitionInstance* fdi =
    CONTAINING_RECORD(fd, CondVD15ns::FilterDefinitionInstance, mDef);
  if (fdi && fdi->prev && fdi->prev->prev) {
    if (&(fdi->prev->next->mDef) != fd) {
      OutputDebugString("-+-+- uuups prev->next != id");
      return 0;
    }
    return &(fdi->prev->mDef);
  }
  return 0;
}

FilterDefinition* CFilterDefinitionHelper::GetNext(FilterDefinition* fd)
{
  if (!m_VD_15x_Mode) return fd->next;

  CondVD15ns::FilterDefinitionInstance* fdi =
    CONTAINING_RECORD(fd, CondVD15ns::FilterDefinitionInstance, mDef);
  if (fdi && fdi->next && fdi->next->next) {
    if (&(fdi->next->prev->mDef) != fd) {
      OutputDebugString("-+-+- uuups next->prev != id");
      return 0;
    }
    return &(fdi->next->mDef);
  }
  return 0;
}

CFilterInfo CFilterDefinitionHelper::GetInfo(FilterDefinition* fd)
{
  CFilterInfo i;

      i.Maker = "-- maker --";
      i.Name  = "-- name --";
      i.Desc  = "-- desc --";

  try {
    try {
      Attach(fd);

      try {
        i.Maker = fd->maker ? fd->maker : "(internal)";
        i.Name  = fd->name;
        i.Desc  = fd->desc;
        i.HasConfig = fd->configProc != 0;
      } __except(1) {
        dout(__FUNC__ " exception");
      }

      Detach(fd);
    } __except(1) {
      dout(__FUNC__ " win exception");
    }
  } catch (...) {
      dout(__FUNC__ " c++ exception");
  }

  return i;
}

void CFilterDefinitionHelper::Attach(FilterDefinition* fd)
{
  if (!m_VD_15x_Mode) return;

  CondVD15ns::FilterDefinitionInstance* fdi =
    CONTAINING_RECORD(fd, CondVD15ns::FilterDefinitionInstance, mDef);

  if (fdi)
    fdi->Attach();
}

void CFilterDefinitionHelper::Detach(FilterDefinition* fd)
{
  if (!m_VD_15x_Mode) return;

  CondVD15ns::FilterDefinitionInstance* fdi =
    CONTAINING_RECORD(fd, CondVD15ns::FilterDefinitionInstance, mDef);

  if (fdi)
    fdi->Detach();
}

FilterDefinition* CFilterDefinitionHelper::FindFilter(const char* Name, FilterDefinition* fd)
{
  CFilterDefinitionHelper it(fd);

  FilterDefinition* Filter = it.GetNext(fd);
  while (Filter && Filter != fd)
  {
    CFilterInfo fi = it.GetInfo(Filter);
    if (!strcmp(fi.Name.c_str(), Name)) return Filter;

    Filter = it.GetNext(Filter);
  }

  Filter = it.GetPrev(fd);
  while (Filter && Filter != fd)
  {
    CFilterInfo fi = it.GetInfo(Filter);
    if (!strcmp(fi.Name.c_str(), Name)) return Filter;

    Filter = it.GetPrev(Filter);
  }

  return 0;
}

CConditionalDetectDeleted::CConditionalDetectDeleted()
{
  m_Count = 0;
  m_MaxCount = 10;
}

void CConditionalDetectDeleted::AddDeleted(void* Ptr)
{
  if (m_Count == m_MaxCount) {
    --m_Count;
    memmove(m_DeletedEntries, m_DeletedEntries + 1, (m_MaxCount - 1)*4);
  }
  m_DeletedEntries[m_Count] = Ptr;
  ++m_Count;
}

bool CConditionalDetectDeleted::CheckDeleted(void* Ptr)
{
  for (int i = 0; i < m_Count; ++i)
    if (m_DeletedEntries[i] == Ptr)
      return true;
  return false;
}

CConditionalDetectDeleted cond_dd;

