%{
//---------------------------------------------------------------------------
//	Conditional Filter for VirtualDub - implements simple branching
//	Copyright (C) 2001-2003 Dmitri Schamschurko
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//      The author can be contacted at:
//      Dmitri Schamschurko
//      dc1@tcs.inf.tu-dresden.de
//
//      Version 1.0     Oktober 2001
//      Version 2.0     March 2003
//---------------------------------------------------------------------------

#include "f_ConditionLogic.h"
#include <malloc.h>

#if defined(__BCPLUSPLUS__)
  #pragma warn -8066 // unreachable code
  #pragma warn -8004 // never used
#endif

int yylex();
int yy_scan_string(const char*);
int yyerror(char* c)
{ return 0; }

extern ExternalFunction GetExternalFunction(const char* Name);

static CChainElement* NewElements = 0;
static CFormula* Formula = 0;
static CNumeric* Numeric = 0;
void Chain(CChainElement* Element)
{
  Element->Next = NewElements;
  NewElements = Element;
}
void Unchain()
{
  CChainElement* Element = NewElements;
  while (Element) {
    CChainElement* ElementNext = Element->Next;
    Element->Next = 0;
    Element = ElementNext;
  }
  NewElements = 0;
}
void FreeElements()
{
  if (NewElements)
    NewElements->Clear(true);
  NewElements = 0;
  Formula = 0;
  Numeric = 0;
}
%}

%union {
  unsigned long val;  /* For returning numbers. */
  const char*   id;   /* For returning strings. */
  CNumeric*     numeric; /* */
  CFormula*     formula; /* */
  NFunction*    function; /* */
}

%token MYEOS
%token LP RP COMMA
%token T_TRUE T_FALSE
%token FRAME STAMP LUMA LUMA601 LUMA709 GRAY
%token BRIGHTNESS LIGHTNESS SATURATION SATURATIONHSB SATURATIONHLS
%token LESS GREATER LESSEQ GREATEREQ
%token T_MIN T_MAX T_AVG RAND
%token <val> NUMBER
%token <id>  T_ID
%left OR
%left AND
%left NOT
%left EQUAL NOTEQ
%left PLUS MINUS
%left TIMES DIV DIVREST
%left UMINUS

%type <formula> exp
%type <numeric> numeric
%type <function> funsymbol
%type <formula>  framelist
%type <formula>  stamplist

%%

input : exp MYEOS { Formula = $1; Numeric = 0; YYACCEPT; }
  | numeric MYEOS { Formula = 0; Numeric = $1; YYACCEPT; }
  | error         { FreeElements(); YYABORT; }
;

exp : LP exp RP             { $$ = new FP($2);   Chain($$); }
  | T_TRUE                  { $$ = new FTrue();  Chain($$); }
  | T_FALSE                 { $$ = new FFalse(); Chain($$); }
  | exp AND exp             { $$ = new FAnd($1,$3);    Chain($$); }
  | exp OR exp              { $$ = new FOr($1,$3);     Chain($$); }
  | NOT exp                 { $$ = new FNot($2);       Chain($$); }
  | exp EQUAL exp           { $$ = new FFEq($1,$3);    Chain($$); }
  | exp NOTEQ exp           { $$ = new FFNotEq($1,$3); Chain($$); }
  | numeric EQUAL numeric           { $$ = new FNEq($1,$3);        Chain($$); }
  | numeric NOTEQ numeric           { $$ = new FNNotEq($1,$3);     Chain($$); }
  | numeric LESS numeric            { $$ = new FNLess($1,$3);      Chain($$); }
  | numeric GREATER numeric         { $$ = new FNGreater($1,$3);   Chain($$); }
  | numeric LESSEQ numeric          { $$ = new FNLessEq($1,$3);    Chain($$); }
  | numeric GREATEREQ numeric       { $$ = new FNGreaterEq($1,$3); Chain($$); }
  | FRAME LP framelist RP           { $$ = $3; }
  | STAMP LP stamplist RP           { $$ = $3; }
;

numeric : NUMBER            { $$ = new NNumber($1);   Chain($$); }
  | LP numeric RP           { $$ = new NP($2);        Chain($$); }
  | numeric PLUS numeric    { $$ = new NPlus($1,$3);  Chain($$); }
  | numeric MINUS numeric   { $$ = new NMinus($1,$3); Chain($$); }
  | numeric TIMES numeric   { $$ = new NTimes($1,$3); Chain($$); }
  | numeric DIV numeric     { $$ = new NDiv($1,$3);     Chain($$); }
  | numeric DIVREST numeric { $$ = new NDivRest($1,$3); Chain($$); }
  | MINUS numeric %prec UMINUS { $$ = new NMinus($2); Chain($$); }
  | T_MIN LP funsymbol RP   { $$ = $3; $3->eModifier = eExMin; }
  | T_MAX LP funsymbol RP   { $$ = $3; $3->eModifier = eExMax; }
  | T_AVG LP funsymbol RP   { $$ = $3; $3->eModifier = eExAvg; }
  | RAND LP numeric RP      { $$ = new NRandom($3);   Chain($$); }
  | funsymbol               { $$ = $1; }
  | FRAME                   { $$ = new NFrame(); Chain($$); }
  | STAMP                   { $$ = new NStamp(); Chain($$); }
;

funsymbol: LUMA             { $$ = new NLuma();       Chain($$); }
  | LUMA601                 { $$ = new NLuma601();    Chain($$); }
  | LUMA709                 { $$ = new NLuma709();    Chain($$); }
  | GRAY                    { $$ = new NGray();       Chain($$); }
  | BRIGHTNESS              { $$ = new NBrightness(); Chain($$); }
  | LIGHTNESS               { $$ = new NLightness();  Chain($$); }
  | SATURATION              { $$ = new NSaturation(); Chain($$); }
  | SATURATIONHSB           { $$ = new NSaturationHSB(); Chain($$); }
  | SATURATIONHLS           { $$ = new NSaturationHLS(); Chain($$); }
  | T_ID                    { ExternalFunction fun = GetExternalFunction($1); if (!fun) {FreeElements(); YYABORT;} $$ = new NFunction($1, fun); Chain($$); }
;

framelist: NUMBER                       { $$ = new FFrame($1);    Chain($$); }
  | NUMBER MINUS NUMBER                 { $$ = new FFrame($1,$3); Chain($$); }
  | framelist COMMA NUMBER              { FFrame* ff = new FFrame($3);    Chain(ff); $$ = new FOr($1, ff); Chain($$); }
  | framelist COMMA NUMBER MINUS NUMBER { FFrame* ff = new FFrame($3,$5); Chain(ff); $$ = new FOr($1, ff); Chain($$); }
;

stamplist: NUMBER                       { $$ = new FStamp($1);    Chain($$); }
  | NUMBER MINUS NUMBER                 { $$ = new FStamp($1,$3); Chain($$); }
  | stamplist COMMA NUMBER              { FStamp* ff = new FStamp($3);    Chain(ff); $$ = new FOr($1, ff); Chain($$); }
  | stamplist COMMA NUMBER MINUS NUMBER { FStamp* ff = new FStamp($3,$5); Chain(ff); $$ = new FOr($1, ff); Chain($$); }
;

%%

#if defined(__BCPLUSPLUS__)
  #pragma warn .8066 // unreachable code
  #pragma warn .8004 // never used
#endif

#ifdef __cplusplus
extern "C" {
#endif
CFormula* parseFormula(const char* pLine)
{
  yy_scan_string(pLine);
  yyparse();
  CFormula* Result = Formula;

  Formula = 0;
  if (Numeric)
    FreeElements();
  else
    Unchain();
  return Result;
}
CNumeric* parseNumeric(const char* pLine)
{
  yy_scan_string(pLine);
  yyparse();
  CNumeric* Result = Numeric;

  Numeric = 0;
  if (Formula)
    FreeElements();
  else
    Unchain();
  return Result;
}
#ifdef __cplusplus
}
#endif

#include <stdio.h>
void maintest()
{
  const char* Line;

  Line = "true";
  yy_scan_string(Line);
  if (yyparse())
    printf ("some error!\n");

  Line = "frame < -2+3 = max(luma) < -3 or \nframe(2 , 3)";
  yy_scan_string(Line);
  if (yyparse())
    printf ("some error!\n");

  Line = "not (2--3<=-----5 and luma < 3)";
  yy_scan_string(Line);
  if (yyparse())
    printf ("some error!\n");

}



